/* $Id: mod_cgi2.c 69 2023-11-26 13:35:57Z nishi $ */
#include <sunhttpd_module.h>
#include <sunhttpd.h>

#include <mtknife.h>

#include <ctype.h>
#include <stdint.h>
#include <sys/wait.h>
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>

const char mod_name[128] = "CGI Module (2)";

bool verbose;

struct entry {
	char* http_path;
	char* realpath;
};

struct entry** entries = NULL;
int entry_count = 0;

void mod_init(bool v){
	verbose = v;
	if(verbose) fprintf(stderr, "CGI2-MOD.. %s got loaded\n", mod_name);
}

void strwrite(struct sunhttpd_module_args args, const char* str){
	args.write(str, strlen(str));
}

int mod_request(struct sunhttpd_module_args args){
	int i;
	bool found = false;
	char* no_query_path = strdup(args.path);
	for(i = 0; no_query_path[i] != 0; i++){
		if(no_query_path[i] == '?'){
			no_query_path[i] = 0;
			break;
		}
	}
	for(i = 0; i < entry_count; i++){	
		if(verbose) fprintf(stderr, "CGI2-MOD.. entry[%d], HTTP Path is %s\n", i, entries[i]->http_path);
		bool runcgi = false;
		runcgi = strcmp(entries[i]->http_path, no_query_path) == 0;
		if(!runcgi){
			char* slash = mtkstrcat(entries[i]->http_path, "/");
			runcgi = strlen(slash) <= strlen(no_query_path) ? strncmp(slash, no_query_path, strlen(slash)) == 0 : false;
			free(slash);
		}
		if(runcgi){
			if(verbose) fprintf(stderr, "CGI2-MOD.. Attempting to run the CGI %s\n", entries[i]->realpath);
			if(access(entries[i]->realpath, X_OK) != 0){
				args.status(500, "Internal Server Error", args.path, args.error_500);
			}else{
				int fds[2];
				pipe(fds);
				pid_t pid = fork();
				/* write to fds[1], read from fds[0] */
				if(pid == 0){
					close(fds[0]);
					close(args.socket);
					dup2(fds[1], 1);
					setenv("SERVER_SOFTWARE", SUNHTTPD_SERVER, 1);
					setenv("PATH_INFO", no_query_path + strlen(entries[i]->http_path), 1);
					setenv("GATEWAY_INTERFACE", "CGI/1.1", 1);
					setenv("REQUEST_METHOD", args.method, 1);
					setenv("SERVER_PROTOCOL", "HTTP/1.1", 1);
					execlp(entries[i]->realpath, entries[i]->realpath, NULL);
					_exit(-1);
				}else{
					close(fds[1]);
					char* buf = malloc(2);
					buf[1] = 0;
					char* line = malloc(1);
					line[0] = 0;
					bool header = true;
					char* headers = malloc(1);
					headers[0] = 0;
					char* status = mtkstrdup("200");
					uint64_t length = 0;
					while(1){
						int r = read(fds[0], buf, 1);
						if(r == 0) break;
						if(header){
							if(buf[0] == '\n'){
								int i;
								bool found = false;
								for(i = 0; line[i] != 0; i++){
									if(line[i] == ':'){
										line[i] = 0;
										if(strcmp(line, "status") == 0){
											int j;
											for(j = i + 1; line[j] != 0; j++){
												if(line[j] != ' '){
													free(status);
													status = mtkstrdup(line + j);
													break;
												}
											}
											found = true;
										}
										line[i] = ':';
										break;
									}else{
										line[i] = tolower(line[i]);
									}
								}
								if(!found){
									char* tmp = headers;
									headers = mtkstrcat3(tmp, line, "\r\n");
									free(tmp);
								}
								if(strlen(line) == 0){
									strwrite(args, "HTTP/1.1 ");
									strwrite(args, status);
									strwrite(args, "\r\n");
									strwrite(args, headers);
									header = false;
								}
								free(line);
								line = malloc(1);
								line[0] = 0;
							}else{
								char* tmp = line;
								line = mtkstrcat(tmp, buf);
								free(tmp);
							}
						}else{
							args.write(buf, 1);
						}
					}
					free(headers);
					free(line);
					free(buf);
					free(status);
					close(fds[0]);
					waitpid(pid, 0, 0);
				}
			}
			free(no_query_path);
			return 0;
		}
	}
	free(no_query_path);
	return 1;
}

int mod_ext(char** args){
	int i;
	for(i = 0; args[i] != NULL; i++);
	if(i < 3){
		if(verbose) fprintf(stderr, "CGI2-MOD.. Need 2 arguments\n");
		printf("CGIPath needs 2 arguments\n");
		return -1;
	}else if(strcmp(args[0], "CGIPath") == 0){
		if(verbose) fprintf(stderr, "CGI2-MOD.. CGIPath, execute %s when %s\n", args[2], args[1]);
		struct entry** new_entries = malloc(sizeof(*new_entries) * (entry_count + 2));
		if(entries != NULL){
			int j;
			for(j = 0; j < entry_count; j++){
				new_entries[j] = entries[j];
			}
			free(entries);
		}
		new_entries[entry_count] = malloc(sizeof(**new_entries));
		new_entries[entry_count]->http_path = strdup(args[1]);
		new_entries[entry_count]->realpath = strdup(args[2]);
		new_entries[entry_count + 1] = NULL;
		entry_count++;
		entries = new_entries;
	}
	return 1;
}
