/* $Id: config.c 69 2023-11-26 13:35:57Z nishi $ */
#include <sunhttpd_config.h>

#include <mtknife.h>

#include <unistd.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <dlfcn.h>

struct mime_entry** entries = NULL;
int entry_count = 0;
struct mime_entry** icon_entries = NULL;
int icon_entry_count = 0;

void** loaded_dl = NULL;
int loaded_dl_count = 0;

char* error_403 = NULL;
char* error_404 = NULL;
char* error_500 = NULL;
int ls_port = 80;
char* root = NULL;
char** denies = NULL;
char* server_root = NULL;
#ifdef ENABLE_SSL
char* ssl_key = NULL;
char* ssl_cert = NULL;
bool do_ssl = false;
#endif

extern bool verbose;

int sunhttpd_load_config(const char* path){
	if(verbose) fprintf(stderr, "CONFIG.... Loading %s\n", path);
	if(server_root == NULL){
		server_root = malloc(PATH_MAX + 1);
		memset(server_root, 0, PATH_MAX + 1);
		getcwd(server_root, PATH_MAX);
	}
	FILE* f = fopen(path, "r");
	if(f == NULL){
		if(verbose) fprintf(stderr, "CONFIG.... Not found\n", path);
		fprintf(stderr, "Config file `%s` was not found\n", path);
		return -1;
	}else{
		char* buf = malloc(2);
		buf[1] = 0;
		char* linebuf = malloc(1);
		linebuf[0] = 0;
		int line = 0;
		int retthis = 0;
		while(1){
			fread(buf, 1, 1, f);
			if(feof(f) || buf[0] == '\n'){
				line++;
				if(strlen(linebuf) > 0){
					if(linebuf[0] != '#'){
						int i;
						bool found = false;
						for(i = 0; linebuf[i] != 0; i++){
							if(linebuf[i] == ' '){
								linebuf[i] = 0;
								found = true;
								break;
							}
						}
						if(found){
							char* key = linebuf;
							char* value = linebuf + i + 1;
							if(strcmp(key, "Port") == 0){
								if(verbose) fprintf(stderr, "CONFIG.... Port is %s\n", value);
								ls_port = atoi(value);
							}else if(strcmp(key, "Root") == 0){
								if(root != NULL) free(root);
								if(verbose) fprintf(stderr, "CONFIG.... Root is %s\n", value);
								root = mtkstrdup(value);
#ifdef ENABLE_SSL
							}else if(strcmp(key, "SSL") == 0){
								if(strcmp(value, "Y") == 0){
									if(verbose) fprintf(stderr, "CONFIG.... TLS/SSL is turned on\n");
									do_ssl = true;
								}else{
									if(verbose) fprintf(stderr, "CONFIG.... TLS/SSL is turned off\n");
									do_ssl = false;
								}
							}else if(strcmp(key, "SSLKey") == 0 && do_ssl){
								if(ssl_key != NULL) free(ssl_key);
								if(verbose) fprintf(stderr, "CONFIG.... TLS/SSL key is %s\n", value);
								ssl_key = mtkstrdup(value);
							}else if(strcmp(key, "SSLCert") == 0 && do_ssl){
								if(ssl_cert != NULL) free(ssl_cert);
								if(verbose) fprintf(stderr, "CONFIG.... TLS/SSL certificate is %s\n", value);
								ssl_cert = mtkstrdup(value);
#endif
							}else if(strcmp(key, "Error403") == 0){
								if(error_403 != NULL) free(error_403);
								error_403 = mtkstrdup(value);
							}else if(strcmp(key, "Error404") == 0){
								if(error_404 != NULL) free(error_404);
								error_404 = mtkstrdup(value);
							}else if(strcmp(key, "Error500") == 0){
								if(error_500 != NULL) free(error_500);
								error_500 = mtkstrdup(value);
							}else if(strcmp(key, "AddMIME") == 0){
								if(entries != NULL){
									struct mime_entry** new_entries = malloc(sizeof(*new_entries) * (entry_count + 2));
									int i;
									for(i = 0; i < entry_count; i++){
										new_entries[i] = entries[i];
										new_entries[i + 1] = NULL;
									}
									free(entries);
									entries = new_entries;
								}else{
									entries = malloc(sizeof(*entries) * 2);
									entries[1] = NULL;
								}
								int i;
								char* mime = NULL;
								char* ext = NULL;
								for(i = 0; value[i] != 0; i++){
									if(value[i] == ' '){
										value[i] = 0;
										ext = mtkstrdup(value);
										mime = mtkstrdup(value + i + 1);
										break;
									}
								}
								if(verbose) fprintf(stderr, "CONFIG.... Added MIME `%s` for the extension `%s`\n", mime, ext);
								entries[entry_count] = malloc(sizeof(**entries));
								entries[entry_count]->ext = ext;
								entries[entry_count]->mime = mime;
								entries[entry_count + 1] = NULL;
								entry_count++;
							}else if(strcmp(key, "AddIcon") == 0 || strcmp(key, "AddIconByName") == 0 || strcmp(key, "AddIconByExt") == 0){
								if(icon_entries != NULL){
									struct mime_entry** new_icon_entries = malloc(sizeof(*new_icon_entries) * (icon_entry_count + 2));
									int i;
									for(i = 0; i < icon_entry_count; i++){
										new_icon_entries[i] = icon_entries[i];
										new_icon_entries[i + 1] = NULL;
									}
									free(icon_entries);
									icon_entries = new_icon_entries;
								}else{
									icon_entries = malloc(sizeof(*icon_entries) * 2);
									icon_entries[1] = NULL;
								}
								int i;
								char* mime = NULL;
								char* ext = NULL;
								for(i = 0; value[i] != 0; i++){
									if(value[i] == ' '){
										value[i] = 0;
										mime = mtkstrdup(value);
										ext = mtkstrdup(value + i + 1);
										break;
									}
								}
								if(verbose) fprintf(stderr, "CONFIG.... Added Icon `%s` for the %s `%s`\n", ext, strcmp(key, "AddIconByName") == 0 ? "Filename" : (strcmp(key, "AddIconByExt") == 0 ? "Extension" : "MIME"), mime);
								icon_entries[icon_entry_count] = malloc(sizeof(**icon_entries));
								icon_entries[icon_entry_count]->ext = ext;
								icon_entries[icon_entry_count]->mime = mime;
								icon_entries[icon_entry_count]->mode = strcmp(key, "AddIconByName") == 0 ? 1 : (strcmp(key, "AddIconByExt") == 0 ? 2 : 0);
								icon_entries[icon_entry_count + 1] = NULL;
								icon_entry_count++;
							}else if(strcmp(key, "Deny") == 0){
								if(denies != NULL){
									int i;
									for(i = 0; denies[i] != NULL; i++) free(denies[i]);
									free(denies);
								}
								int i;
								int start = 0;
								int count = 0;
								for(i = 0;; i++){
									if(value[i] == ';' || value[i] == 0){
										count++;
										if(value[i] == 0) break;
									}
								}
								denies = malloc(sizeof(*denies) * (count + 1));
								denies[count] = NULL;
								int putinto = 0;
								for(i = 0;; i++){
									if(value[i] == ';' || value[i] == 0){
										char* deny = malloc(i - start + 1);
										deny[i - start] = 0;
										memcpy(deny, value + start, i - start);
										denies[putinto] = deny;
										if(verbose) fprintf(stderr, "CONFIG.... Denying %s\n", deny);
										putinto++;
										start = i + 1;
										if(value[i] == 0) break;
									}
								}
							}else if(strcmp(key, "LoadModule") == 0){
								chdir(server_root);
								void** new_loaded_dl;
								new_loaded_dl = malloc(sizeof(*new_loaded_dl) * (loaded_dl_count + 2));
								if(loaded_dl != NULL){
									int i;
									for(i = 0; i < loaded_dl_count; i++){
										new_loaded_dl[i] = loaded_dl[i];
									}
								}
								void* handler = dlopen(value, RTLD_LAZY);
								new_loaded_dl[loaded_dl_count] = handler;
								new_loaded_dl[loaded_dl_count + 1] = NULL;
								free(loaded_dl);
								if(handler == NULL){
									if(verbose) fprintf(stderr, "MODULE.... Error %s\n", dlerror());
									fprintf(stderr, "Error: %s: Line %d: Failed to load the module.\n", path, line);
									retthis = -1;
								}else{
									if(verbose){
										const char* name = dlsym(handler, "mod_name");
										void(*mod_init)(bool) = dlsym(handler, "mod_init");
										mod_init(verbose);
										fprintf(stderr, "MODULE.... Loaded the module %s, name `%s`\n", value, name);
									}
								}
								loaded_dl = new_loaded_dl;
								loaded_dl_count++;
							}else if(strcmp(key, "External") == 0){
								int i;
								char** args = malloc(sizeof(*args));
								args[0] = NULL;
								char* str = malloc(1);
								str[0] = 0;
								char* buf = malloc(2);
								buf[1] = 0;
								int arg_counter = 0;
								for(i = 0;; i++){
									buf[0] = value[i];
									if(value[i] == ' ' || value[i] == 0){
										if(strlen(str) > 0){
											char** new_args = malloc(sizeof(*args) * (arg_counter + 2));
											int j;
											for(j = 0; j < arg_counter; j++){
												new_args[j] = args[j];
											}
											free(args);
											new_args[arg_counter] = str;
											new_args[arg_counter + 1] = NULL;
											args = new_args;
											arg_counter++;
										}
										str = malloc(1);
										str[0] = 0;
										if(value[i] == 0) break;
									}else{
										char* tmp = str;
										str = mtkstrcat(tmp, buf);
										free(tmp);
									}
								}
								free(buf);
								for(i = 0; loaded_dl[i] != NULL; i++){
									int(*mod_ext)(char**) = dlsym(loaded_dl[i], "mod_ext");
									if(mod_ext != NULL){
										int st = mod_ext(args);
										if(st == 0){
											break;
										}else if(st == -1){
											retthis = -1;
											break;
										}
									}
								}
								for(i = 0; args[i] != NULL; i++) free(args[i]);
								free(args);
							}else if(strcmp(key, "ServerRoot") == 0){
								free(server_root);
								server_root = mtkstrdup(value);
							}else{
								if(verbose) fprintf(stderr, "CONFIG.... Invalid directive at line %d\n", line);
								fprintf(stderr, "Error: %s: Line %d: Invalid.\n", path, line);
								retthis = -1;
							}
						}else{
							if(verbose) fprintf(stderr, "CONFIG.... Invalid line at line %d\n", line);
							fprintf(stderr, "Error: %s: Line %d: Invalid.\n", path, line);
							retthis = -1;
						}
					}
				}
				free(linebuf);
				linebuf = malloc(1);
				linebuf[0] = 0;
				if(feof(f)) break;
			}else{
				char* tmp = linebuf;
				linebuf = mtkstrcat(tmp, buf);
				free(tmp);
			}
		}
		free(linebuf);
		free(buf);
		fclose(f);
#ifdef ENABLE_SSL
		if(do_ssl){
			if(ssl_key == NULL){
				fprintf(stderr, "Error: %s: SSL was turned on, but SSLKey was not set\n", path);
				retthis = -1;
			}
			if(ssl_cert == NULL){
				fprintf(stderr, "Error: %s: SSL was turned on, but SSLCert was not set\n", path);
				retthis = -1;
			}
		}
#endif
		return retthis;
	}
}
