/* $Id: server.c 68 2023-11-25 15:08:39Z nishi $ */
#include <sunhttpd_server.h>
#include <sunhttpd_config.h>
#include <sunhttpd_module.h>
#include <sunhttpd.h>

#include <ls_server.h>
#include <mtknife.h>

#include <ctype.h>
#include <stdio.h>
#include <stdbool.h>
#include <unistd.h>
#include <string.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <dirent.h>
#include <stdlib.h>

#include <dlfcn.h>

#ifdef ENABLE_SSL
#include <openssl/ssl.h>
#include <openssl/err.h>
SSL_CTX* ctx;
#endif

#define BUFSIZE 4096	/* HTTP write buffer size */
#define BUFFER_SIZE 256	/* HTTP read buffer size */

extern void** loaded_dl;
extern struct mime_entry** entries;
extern struct mime_entry** icon_entries;

extern int ls_port;
extern char* root;
extern bool verbose;
extern char** denies;
extern char* error_403;
extern char* error_404;
extern char* error_500;

int fd;
#ifdef ENABLE_SSL
SSL* ssl;
extern bool do_ssl;
extern char* ssl_key;
extern char* ssl_cert;
#endif

char* sunhttpd_get_ext(const char* path_){
	bool found = false;
	int st = -1;
	char* path = mtkstrdup(path_);
	int i;
	for(i = 0; path_[i] != 0; i++){
		if(path_[i] == '.'){
			path[i] = 0;
			if(i != strlen(path_) - 1) found = true;
			st = i;
		}
	}
	char* str = mtkstrdup(path + (st == -1 ? 0 : st + 1));
	free(path);
	return str;
}

ssize_t sunhttpd_read(void* buf, size_t count){
#ifdef ENABLE_SSL
	if(do_ssl){
		return SSL_read(ssl, buf, count);
	}else{
		return read(fd, buf, count);
	}
#else
	return read(fd, buf, count);
#endif
}

ssize_t sunhttpd_write(const void* buf, size_t count){
#ifdef ENABLE_SSL
	if(!do_ssl){
		return write(fd, buf, count);
	}else{
		return SSL_write(ssl, buf, count);
	}
#else
	return write(fd, buf, count);
#endif
}

#define sunhttpd_string_write(x) sunhttpd_write(x, strlen(x))

char* sunhttpd_get_mime(const char* path_){
	char* mime = mtkstrdup("application/octet-stream");
	char* path = mtkstrdup(path_);
	int i;
	bool found = false;
	if(verbose) fprintf(stderr, "MIME...... %s\n", path_);
	int st;
	for(i = 0; path_[i] != 0; i++){
		if(path_[i] == '.'){
			path[i] = 0;
			if(i != strlen(path_) - 1) found = true;
			st = i;
		}
	}
	if(found){
		int j;
		for(j = 0; entries[j] != NULL; j++){
			if(strcmp(entries[j]->ext, path + st + 1) == 0){
				free(mime);
				mime = mtkstrdup(entries[j]->mime);
				break;
			}
		}
	}
	if(verbose) fprintf(stderr, "MIME...... %s, %s\n", path_, mime);
	free(path);
	return mime;
}

void return_status(int st, const char* text, const char* reqpath, const char* path){
	if(verbose) fprintf(stderr, "STATUS.... %d\n", st);
	if(st == 200){
		char* p = mtkstrcat(root, reqpath);
		if(verbose) fprintf(stderr, "REALPATH.. %s\n", p);
		struct stat s;
		if(stat(p, &s) == 0){
			if(verbose) fprintf(stderr, "DIRECTORY. %s\n", S_ISDIR(s.st_mode) ? "Yes" : "No");
			if(S_ISDIR(s.st_mode)){
				if(p[strlen(p) - 1] != '/'){
					if(verbose) fprintf(stderr, "REDIRECT.. Path was not ended with slash\n");
					if(verbose) fprintf(stderr, "REDIRECT.. Going to %s/\n", reqpath);
					sunhttpd_string_write("HTTP/1.1 301 Moved Permanently\r\n");
					sunhttpd_write("Server: ", 8);
					sunhttpd_write(SUNHTTPD_SERVER, strlen(SUNHTTPD_SERVER));
					sunhttpd_write("\r\n", 2);
					sunhttpd_string_write("Location: ");
					sunhttpd_string_write(reqpath);
					sunhttpd_string_write("/\r\n");
					sunhttpd_write("\r\n", 2);
					sunhttpd_write("\r\n", 2);
				}else{
					char* index = mtkstrcat(p, "/index.html");
					if(stat(index, &s) == 0){
						sunhttpd_write("HTTP/1.1 ", 9);
						char* code = malloc(4);
						sprintf(code, "%d", st);
						code[3] = 0;
						sunhttpd_write(code, 3);
						free(code);
						sunhttpd_write(" ", 1);
						sunhttpd_write(text, strlen(text));
						sunhttpd_write("\r\n", 2);
						sunhttpd_write("Server: ", 8);
						sunhttpd_write(SUNHTTPD_SERVER, strlen(SUNHTTPD_SERVER));
						sunhttpd_write("\r\n", 2);
						char* mime = sunhttpd_get_mime(index);
						sunhttpd_write("Content-Type: ", 14);
						sunhttpd_write(mime, strlen(mime));
						free(mime);
						sunhttpd_write("\r\n", 2);
						char* len = malloc(513);
						memset(len, 0, 513);
						sprintf(len, "%ld", s.st_size);
						sunhttpd_write("Content-Length: ", 16);
						sunhttpd_write(len, strlen(len));
						free(len);
						sunhttpd_write("\r\n", 2);
						sunhttpd_write("\r\n", 2);
						char* buf = malloc(BUFSIZE);
						FILE* f = fopen(index, "r");
						while(1){
							int a = fread(buf, 1, BUFSIZE, f);
							if(a == 0) break;
							sunhttpd_write(buf, a);
						}
						fclose(f);
					}else{
						struct dirent** namelist;
						int n = scandir(p, &namelist, NULL, alphasort);
						if(n >= 0){
							char* resp = mtkstrdup("<!DOCTYPE html><html><head><meta charset=\"UTF-8\">");
							char* tmp;
							char* title = mtkstrcat3("<title>Index of ", reqpath, "</title>");
							tmp = resp;
							resp = mtkstrcat(resp, title);
							free(tmp);
							tmp = resp;
							resp = mtkstrcat3(resp, "</head><body><h1>Index of ", reqpath);
							free(tmp);
							tmp = resp;
							resp = mtkstrcat(resp, "</h1><hr><table width=\"100%\"><tr><th align=\"left\">Name</th><th align=\"left\">MIME</th><th align=\"left\">Size</th></tr>");
							free(tmp);
							int i;
#define PROC(x)						for(i = 0; i < n; i++){ \
								if((DT_DIR & namelist[i]->d_type) == x) continue; \
								if(strcmp(namelist[i]->d_name, ".") != 0){ \
									tmp = resp; \
									resp = mtkstrcat(resp, "<tr><td>"); \
									free(tmp); \
									bool found_it = false; \
									int j; \
									char* p2 = mtkstrcat3(p, "/", namelist[i]->d_name); \
									stat(p2, &s); \
									if(S_ISDIR(s.st_mode)){ \
										for(j = 0; icon_entries[j] != NULL; j++){ \
											if(icon_entries[j]->mode != 0) continue; \
											if(strcmp("misc/dir", icon_entries[j]->mime) == 0){ \
												found_it = true; \
											} \
											if(found_it) break; \
										} \
										if(found_it){ \
											tmp = resp; \
											resp = mtkstrcat3(resp, "<img src=\"", icon_entries[j]->ext); \
											free(tmp); \
											tmp = resp; \
											resp = mtkstrcat(resp, "\"> "); \
											free(tmp); \
										} \
									}else{ \
										for(j = 0; icon_entries[j] != NULL; j++){ \
											if(icon_entries[j]->mode != 2) continue; \
											char* ext = sunhttpd_get_ext(namelist[i]->d_name); \
											if(strcmp(ext, icon_entries[j]->mime) == 0){ \
												found_it = true; \
											} \
											free(ext); \
											if(found_it) break; \
										} \
										if(found_it){ \
											tmp = resp; \
											resp = mtkstrcat3(resp, "<img src=\"", icon_entries[j]->ext); \
											free(tmp); \
											tmp = resp; \
											resp = mtkstrcat(resp, "\"> "); \
											free(tmp); \
											goto go_free ## x; \
										} \
 \
										for(j = 0; icon_entries[j] != NULL; j++){ \
											if(icon_entries[j]->mode != 1) continue; \
											if(strcmp(namelist[i]->d_name, icon_entries[j]->mime) == 0){ \
												found_it = true; \
											} \
											if(found_it) break; \
										} \
										if(found_it){ \
											tmp = resp; \
											resp = mtkstrcat3(resp, "<img src=\"", icon_entries[j]->ext); \
											free(tmp); \
											tmp = resp; \
											resp = mtkstrcat(resp, "\"> "); \
											free(tmp); \
											goto go_free ## x; \
										} \
 \
										char* mimetype = sunhttpd_get_mime(namelist[i]->d_name); \
										for(j = 0; icon_entries[j] != NULL; j++){ \
											if(icon_entries[j]->mode != 0) continue; \
											if(strcmp(mimetype, icon_entries[j]->mime) == 0){ \
												found_it = true; \
											} \
											if(found_it) break; \
										} \
										free(mimetype); \
										if(found_it){ \
											tmp = resp; \
											resp = mtkstrcat3(resp, "<img src=\"", icon_entries[j]->ext); \
											free(tmp); \
											tmp = resp; \
											resp = mtkstrcat(resp, "\"> "); \
											free(tmp); \
										} \
go_free ## x:; \
									} \
									free(p2); \
									tmp = resp; \
									resp = mtkstrcat3(resp, "<a href=\"", namelist[i]->d_name); \
									free(tmp); \
									tmp = resp; \
									resp = mtkstrcat3(resp, "\">", namelist[i]->d_name); \
									free(tmp); \
									tmp = resp; \
									char* len = malloc(513); \
									memset(len, 0, 513); \
									char* pth = mtkstrcat3(p, "/", namelist[i]->d_name); \
									stat(pth, &s); \
									free(pth); \
									if(S_ISDIR(s.st_mode)){ \
										resp = mtkstrcat(resp, "/</a></td><td></td><td>&lt;DIR&gt;"); \
										free(tmp); \
									}else{ \
										sprintf(len, "%ld", s.st_size); \
										resp = mtkstrcat(resp, "</a></td><td>"); \
										free(tmp); \
 \
										char* mime = sunhttpd_get_mime(namelist[i]->d_name); \
										tmp = resp; \
										resp = mtkstrcat(resp, mime); \
										free(tmp); \
										free(mime); \
 \
										tmp = resp; \
										resp = mtkstrcat3(resp, "</td><td>", len); \
										free(tmp); \
										free(len); \
									} \
									tmp = resp; \
									resp = mtkstrcat(resp, "</td></tr>"); \
									free(tmp); \
								} \
							}
							PROC(0);
							PROC(DT_DIR);
							for(i = 0; i < n; i++) free(namelist[i]);
							free(namelist);
							tmp = resp;
							resp = mtkstrcat(resp, "</table><hr>");
							free(tmp);
							char* readme = mtkstrcat(p, "/README");
							if(stat(readme, &s) == 0){
								char* buf = malloc(s.st_size + 1);
								buf[s.st_size] = 0;
								FILE* f = fopen(readme, "r");
								fread(buf, s.st_size, 1, f);
								fclose(f);
								tmp = resp;
								resp = mtkstrcat3(resp, "<pre>", buf);
								free(tmp);
								tmp = resp;
								resp = mtkstrcat(resp, "</pre><hr>");
								free(tmp);
								free(buf);
							}
							free(readme);
							tmp = resp;
							resp = mtkstrcat3(resp, "<i>", SUNHTTPD_TEXT);
							free(tmp);
							tmp = resp;
							resp = mtkstrcat(resp, "</i>");
							free(tmp);
							tmp = resp;
							resp = mtkstrcat(resp, "</body></html>");
							free(tmp);
							sunhttpd_write("HTTP/1.1 ", 9);
							char* code = malloc(4);
							sprintf(code, "%d", st);
							code[3] = 0;
							sunhttpd_write(code, 3);
							free(code);
							sunhttpd_write(" ", 1);
							sunhttpd_write(text, strlen(text));
							sunhttpd_write("\r\n", 2);
							sunhttpd_string_write("Content-Type: text/html\r\n");
							char* len = malloc(513);
							memset(len, 0, 513);
							sprintf(len, "%ld", strlen(resp));
							sunhttpd_write("Content-Length: ", 16);
							sunhttpd_write(len, strlen(len));
							free(len);
							sunhttpd_write("\r\n", 2);
							sunhttpd_write("\r\n", 2);
							sunhttpd_string_write(resp);
							free(resp);
						}
					}
					free(index);
				}
			}else{
				sunhttpd_write("HTTP/1.1 ", 9);
				char* code = malloc(4);
				sprintf(code, "%d", st);
				code[3] = 0;
				sunhttpd_write(code, 3);
				sunhttpd_write(" ", 1);
				sunhttpd_write(text, strlen(text));
				sunhttpd_write("\r\n", 2);
				sunhttpd_write("Server: ", 8);
				sunhttpd_write(SUNHTTPD_SERVER, strlen(SUNHTTPD_SERVER));
				sunhttpd_write("\r\n", 2);
				char* mime = sunhttpd_get_mime(reqpath);
				sunhttpd_write("Content-Type: ", 14);
				sunhttpd_write(mime, strlen(mime));
				free(mime);
				sunhttpd_write("\r\n", 2);
				char* len = malloc(513);
				memset(len, 0, 513);
				sprintf(len, "%ld", s.st_size);
				sunhttpd_write("Content-Length: ", 16);
				sunhttpd_write(len, strlen(len));
				free(len);
				sunhttpd_write("\r\n", 2);
				sunhttpd_write("\r\n", 2);
				char* buf = malloc(512);
				FILE* f = fopen(p, "r");
				while(1){
					int a = fread(buf, 1, 512, f);
					if(a == 0) break;
					sunhttpd_write(buf, a);
				}
				fclose(f);
			}
		}else{
			if(verbose) fprintf(stderr, "ERROR..... stat() failed\n");
		}
		free(p);
	}else{
		char* p = mtkstrcat(root, path);
		struct stat s;
		sunhttpd_write("HTTP/1.1 ", 9);
		char* code = malloc(4);
		sprintf(code, "%d", st);
		code[3] = 0;
		sunhttpd_write(code, 3);
		free(code);
		sunhttpd_write(" ", 1);
		sunhttpd_write(text, strlen(text));
		sunhttpd_write("\r\n", 2);
		sunhttpd_write("Server: ", 8);
		sunhttpd_write(SUNHTTPD_SERVER, strlen(SUNHTTPD_SERVER));
		sunhttpd_write("\r\n", 2);
		if(stat(p, &s) == 0){
			char* mime = sunhttpd_get_mime(p);
			sunhttpd_write("Content-Type: ", 14);
			sunhttpd_write(mime, strlen(mime));
			free(mime);
			sunhttpd_write("\r\n", 2);
			char* len = malloc(513);
			memset(len, 0, 513);
			sprintf(len, "%ld", s.st_size);
			sunhttpd_write("Content-Length: ", 16);
			sunhttpd_write(len, strlen(len));
			free(len);
			sunhttpd_write("\r\n", 2);
			sunhttpd_write("\r\n", 2);
			char* buf = malloc(512);
			FILE* f = fopen(p, "r");
			while(1){
				int a = fread(buf, 1, 512, f);
				if(a == 0) break;
				sunhttpd_write(buf, a);
			}
			fclose(f);
		}else{
			if(verbose) fprintf(stderr, "ERROR..... stat() failed :(");
			sunhttpd_string_write("Content-Type: text/plain\r\n");
			char* len = malloc(513);
			memset(len, 0, 513);
			sprintf(len, "%ld", strlen(text) + 2);
			sunhttpd_write("Content-Length: ", 16);
			sunhttpd_write(len, strlen(len));
			free(len);
			sunhttpd_string_write("\r\n\r\n");
			sunhttpd_string_write(text);
			sunhttpd_string_write("\r\n");
		}
		free(p);
	}
}

char gethex(const char n){
	if(n >= '0' && n <= '9') return n - '0';
	if(n >= 'a' && n <= 'f') return n - 'a';
	if(n >= 'A' && n <= 'F') return n - 'A';
	return 0;
}

void sock_handl(int sock){
	int yes = 1;
	setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, &yes, sizeof(yes));
	fd = sock;
#ifdef ENABLE_SSL
	if(do_ssl){
		ssl = SSL_new(ctx);
		SSL_set_fd(ssl, fd);
		SSL_accept(ssl);
	}
#endif
	char ch;
	char* linebuf;
	int line;
	char* method = NULL;
	char* path = NULL;
	bool pathtrv = false;
	bool repeat_it = false;
	char http_buf[BUFFER_SIZE];
repeat_again:;
	repeat_it = false;
	linebuf = malloc(1);
	linebuf[0] = 0;
	line = 0;
	method = NULL;
	path = NULL;
	pathtrv = false;
	while(1){
		int counter = sunhttpd_read(http_buf, BUFFER_SIZE);
		if(counter < 0) break;
		int buf_count;
		bool too_bad = false;
		for(buf_count = 0; buf_count < counter; buf_count++){
			ch = http_buf[buf_count];
			if(ch == '\n'){
				bool empty = strlen(linebuf) == 0;
				if(!empty){
					if(line == 0){
						int i;
						bool found = false;
						for(i = 0; linebuf[i] != 0; i++){
							if(linebuf[i] == ' '){
								linebuf[i] = 0;
								found = true;
								break;
							}
						}
						if(found){
							method = mtkstrdup(linebuf);
						}
	
						found = false;
						int j;
						for(j = i + 1; linebuf[j] != 0; j++){
							if(linebuf[j] == ' '){
								linebuf[j] = 0;
								found = true;
								break;
							}
						}
						if(found){
							path = mtkstrdup(linebuf + i + 1);
							if(path[0] != '/'){
								/* Avoid the path traversal */
								pathtrv = true;
								break;
							}
						}
					}else{
						if(verbose) fprintf(stderr, "REQUEST... %s\n", linebuf);
						int i;
						for(i = 0; linebuf[i] != 0; i++){
							if(linebuf[i] == ':'){
								int j;
								for(j = i; linebuf[j] != 0; j++){
									if((j == i && linebuf[j] == ':') || linebuf[j] == '\t' || linebuf[j] == ' '){	
										linebuf[j] = 0;
									}else{
										break;
									}
								}
								int k;
								for(k = 0; linebuf[k] != 0; k++){
									linebuf[k] = tolower(linebuf[k]);
								}
								if(verbose) fprintf(stderr, "REQUEST... Name `%s`\n", linebuf);
								if(verbose) fprintf(stderr, "REQUEST... Data `%s`\n", linebuf + j);
								if(strcmp(linebuf, "connection") == 0){
									if(strcmp(linebuf + j, "keep-alive") == 0){
										repeat_it = true;
									}
								}
							}
						}
					}
					line++;
				}
				free(linebuf);
				linebuf = malloc(1);
				linebuf[0] = 0;
				if(empty){
					too_bad = true;
					break;
				}
			}else if(ch != '\r'){
				char* buf = malloc(2);
				buf[0] = ch;
				buf[1] = 0;
				char* tmp = linebuf;
				linebuf = mtkstrcat(tmp, buf);
				free(tmp);
				free(buf);
			}
		}
		if(buf_count < BUFFER_SIZE){
			break;
		}
	}
	free(linebuf);
	char* new_path = malloc(1);
	new_path[0] = 0;
	int i;
	char* buf = malloc(2);
	buf[1] = 0;
	int counter = -1;
	char ch2 = 0;
	for(i = 0; path[i] != 0; i++){
		buf[0] = path[i];
		bool pdetect = true;
		if(counter >= 0){
			char a = gethex(buf[0]);
			ch2 = ch2 << 4;
			ch2 |= (char)a;
			counter++;
			if(counter == 2){
				counter = -1;
				pdetect = false;
				buf[0] = ch2;
				ch2 = 0;
			}else{
				continue;
			}
		}
		if(buf[0] == '%' && pdetect){
			counter = 0;
		}else{
			char* tmp = new_path;
			new_path = mtkstrcat(tmp, buf);
			free(tmp);
		}
	}
	free(path);
	path = new_path;
	free(buf);
	if(method != NULL && verbose){
		fprintf(stderr, "METHOD.... %s\n", method);
	}
	if(path != NULL && verbose){
		fprintf(stderr, "PATH...... %s\n", path);
	}
	if(pathtrv){
		free(path);
		path = mtkstrdup("/");
	}else{
		char rootdir[PATH_MAX + 1];
		chdir(root);
		memset(rootdir, 0, PATH_MAX + 1);
		getcwd(rootdir, PATH_MAX);
		char* new_rootdir = mtkstrcat(rootdir, "/");
		chdir(root);
		char* new_path = mtkstrdup(path);
		int st = 0;
		for(i = 0;; i++){
			if(new_path[i] == '/' || new_path[i] == 0){
				bool decide = new_path[i] == 0;
				new_path[i] = 0;
				if(strlen(new_path + st) > 0){
					chdir(new_path + st);
				}
				st = i + 1;
				if(decide) break;
			}
		}
		free(new_path);
		char dir[PATH_MAX + 1];
		memset(dir, 0, PATH_MAX + 1);
		getcwd(dir, PATH_MAX);
		char* n_dir;
		if(dir[strlen(dir) - 1] == '/'){
			n_dir = mtkstrdup(dir);
		}else{
			n_dir = mtkstrcat(dir, "/");
		}
		if(strncmp(rootdir, n_dir, strlen(rootdir)) != 0) pathtrv = true;
		free(n_dir);
		if(pathtrv){
			free(path);
			path = mtkstrdup("/");
		}
		free(new_rootdir);
	}
	char* p = mtkstrcat(root, path);
	char* npath = mtkstrdup(path);
	for(i = 0; npath[i] != 0; i++){
		if(npath[i] == '?'){
			npath[i] = 0;
			break;
		}
	}
	for(i = 0; p[i] != 0; i++){
		if(p[i] == '?'){
			p[i] = 0;
			break;
		}
	}
	bool retnow = false;
	if(loaded_dl != NULL){
		for(i = 0; loaded_dl[i] != NULL; i++){
			if(verbose) fprintf(stderr, "MODULE.... Trying the module %d\n", i);
			int(*mod_request)(struct sunhttpd_module_args) = dlsym(loaded_dl[i], "mod_request");
			if(mod_request != NULL){
				struct sunhttpd_module_args args;
				args.write = sunhttpd_write;
				args.read = sunhttpd_read;
				args.path = path;
				args.verbose = verbose;
				args.method = method;
				args.error_403 = error_403;
				args.error_404 = error_404;
				args.error_500 = error_500;
				args.status = return_status;
				args.socket = sock;
				int rt = mod_request(args);
				if(rt == 0){
					if(verbose) fprintf(stderr, "MODULE.... Module returned 0, breaking\n");
					retnow = true;
					break;
				}
			}
		}
	}
	if(!retnow){
		struct stat s;
		if(stat(p, &s) != 0){
			return_status(404, "Not Found", npath, error_404);
		}else{
			bool denyacc = false;
			if(denies != NULL){
				for(i = 0; denies[i] != NULL; i++){
					if(strcmp(denies[i], path) == 0){
						denyacc = true;
						break;
					}
				}
			}
			if(denyacc){
				fprintf(stderr, "STATUS.... Forbidden\n");
				return_status(403, "Forbidden", npath, error_403);
			}else{
				return_status(200, "OK", npath, NULL);
			}
		}
	}
	free(npath);
	free(p);
	if(method != NULL) free(method);
	if(path != NULL) free(path);
#ifdef ENABLE_SSL
	if(do_ssl){
		SSL_shutdown(ssl);
		SSL_free(ssl);
	}
#endif
}

int sunhttpd_server(void){
#ifdef ENABLE_SSL
	if(do_ssl){
		SSL_load_error_strings();
		OpenSSL_add_ssl_algorithms();
		ctx = SSL_CTX_new(TLS_server_method());
		if(ctx == NULL){
			fprintf(stderr, "SSL Error.\n");
			ERR_print_errors_fp(stderr);
			return -1;
		}
		SSL_CTX_set_ecdh_auto(ctx, 1);
		if(SSL_CTX_use_PrivateKey_file(ctx, ssl_key, SSL_FILETYPE_PEM) <= 0){
			fprintf(stderr, "Failed to load SSL private key\n");
			return -1;
		}
		if(SSL_CTX_use_certificate_file(ctx, ssl_cert, SSL_FILETYPE_PEM) <= 0){
			fprintf(stderr, "Failed to load SSL certificate\n");
			return -1;
		}
	}
#endif
	if(ls_start_server(ls_port, sock_handl) == -1){
		fprintf(stderr, "Failed to start the server\n");
		return -1;
	}
#ifdef ENABLE_SSL
	if(do_ssl){
		SSL_CTX_free(ctx);
		EVP_cleanup();
	}
#endif
}
