/* $Id: server.c 14 2023-11-27 14:19:34Z nishi $ */
#define LS_SERVER_SRC
#include <ls_server.h>

#include <stdlib.h>
#include <stdio.h>

#include <sys/socket.h>
#include <netinet/tcp.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <stdbool.h>

bool able_fd = true;
struct fd_info* fds = NULL;

int* my_pipes;
int* my_pipes2;

/* Free this yourself! It's not even statically allocated. */
int* ls_get_sockets(void){
	kill(getppid(), SIGUSR2);
	int length;
	read(my_pipes[0], &length, sizeof(length)); /* Length of the fd_info */
	int* infos = malloc(sizeof(*infos) * (length + 1));
	infos[length] = -1;
	int i = 0;
	for(i = 0; i < length; i++){
		read(my_pipes[0], &infos[i], sizeof(*infos));
	}
	return infos;
}

ssize_t ls_write(int sockfd, void* buf, size_t len){
	write(my_pipes2[1], &sockfd, sizeof(sockfd));
	write(my_pipes2[1], &len, sizeof(len));
	write(my_pipes2[1], buf, len);
	kill(getppid(), SIGRTMIN + 4);
	ssize_t size;
	read(my_pipes[0], &size, sizeof(size));
	return size;
}

void ls_kill_fd(int sig, siginfo_t* info, void* ctx){
	if(sig == SIGUSR2){
		/* Child requested to send the stuff */
		while(!able_fd);
		able_fd = false;
		int i = 0;
		for(i = 0; fds[i].fd != -1; i++);
		i = 0;
		int* pipes = NULL;
		while(1){
			if(fds[i].fd == -1) break;
			if(fds[i].pid == info->si_pid){
				pipes = fds[i].pipes;
			}
			i++;
		}
		if(pipes != NULL){
			write(pipes[1], &i, sizeof(i));
			int count = i;
			for(i = 0; i < count; i++){
				write(pipes[1], &fds[i].fd, sizeof(fds[i].fd));
			}
		}
		able_fd = true;
	}else if(sig == SIGRTMIN + 4){
		while(!able_fd);
		able_fd = false;
		int i = 0;
		int* pipes = NULL;
		int* pipes2 = NULL;
		while(1){
			if(fds[i].fd == -1) break;
			if(fds[i].pid == info->si_pid){
				pipes = fds[i].pipes;
				pipes2 = fds[i].pipes2;
			}
			i++;
		}
		if(pipes != NULL){
			int sockfd;
			read(pipes2[0], &sockfd, sizeof(sockfd));
			size_t len;
			read(pipes2[0], &len, sizeof(len));
			char* buf = malloc(len);
			read(pipes2[0], buf, len);
			ssize_t size = write(sockfd, buf, len);
			free(buf);
			write(pipes[1], &size, sizeof(size));
		}
		able_fd = true;
	}else{
		while(!able_fd);
		able_fd = false;
		int i = 0;
		struct fd_info* new_fds = NULL;
		int counter = 0;
		while(1){
			if(fds[i].fd == -1) break;
			if(fds[i].pid == info->si_pid){
				int j;
				close(fds[i].fd);
				fds[i].fd = -2; /* Mark this as dead */
			}else{
				counter++;
			}
			i++;
		}
		new_fds = malloc(sizeof(*new_fds) * (counter + 1));
		new_fds[counter].fd = -1;
		int j;
		int counter2 = 0;
		for(j = 0; fds[j].fd != -1; j++){
			if(fds[j].fd == -2) continue;
			new_fds[counter2] = fds[j];
			counter2++;
		}
		if(new_fds != NULL){
			free(fds);
			fds = new_fds;
		}
		able_fd = true;
	}
}

bool ls_wait;

void unlock_wait(int sig){
	ls_wait = false;
}

int ls_start_server(int port, handler socket_handler){
	if(fds == NULL){
		fds = malloc(sizeof(*fds));
		fds[0].fd = -1;
	}
	int server_socket;
	struct sockaddr_in6 server_address;
	if((server_socket = socket(AF_INET6, SOCK_STREAM, IPPROTO_TCP)) < 0){
		return -1;
	}
	int yes = 1;
	if(setsockopt(server_socket, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes)) < 0){
		close(server_socket);
		return -1;
	}
	if(setsockopt(server_socket, IPPROTO_TCP, TCP_NODELAY, &yes, sizeof(yes)) < 0){
		close(server_socket);
		return 1;
	}
	int no = 0;
	if(setsockopt(server_socket, IPPROTO_IPV6, IPV6_V6ONLY, &no, sizeof(no)) < 0){
		close(server_socket);
		return 1;
	}
	memset(&server_address, 0, sizeof(server_address));
	server_address.sin6_family = AF_INET6;
	server_address.sin6_addr = in6addr_any;
	server_address.sin6_port = htons(port);
	if(bind(server_socket, (struct sockaddr*)&server_address, sizeof(server_address)) < 0){
		close(server_socket);
		return -1;
	}
	if(listen(server_socket, 128) < 0){
		close(server_socket);
		return -1;
	}
	struct sigaction sa;
	memset(&sa, 0, sizeof(sa));
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_SIGINFO | SA_RESTART;
	sa.sa_sigaction = ls_kill_fd;
	sigaction(SIGCHLD, &sa, NULL);
	sigaction(SIGUSR2, &sa, NULL);
	sigaction(SIGRTMIN + 4, &sa, NULL);
	while(1){
		struct sockaddr_in claddr;
		int clen = sizeof(claddr);
		int sock = accept(server_socket, (struct sockaddr*)&claddr, &clen);
		ls_wait = true;
		signal(SIGUSR1, unlock_wait);
		int* pipes = malloc(sizeof(*pipes) * 2);
		pipe(pipes);
		int* pipes2 = malloc(sizeof(*pipes2) * 2);
		pipe(pipes2);
		pid_t p = fork();
		if(p == 0){
			while(ls_wait);
			my_pipes = pipes;
			my_pipes2 = pipes2;
			if(socket_handler == NULL){
				dprintf(sock, "Socket handler was not set. Goodbye\n");
			}else{
				socket_handler(sock);
			}
			_exit(0);
		}else{
			while(!able_fd);
			able_fd = false;
			int i;
			for(i = 0; fds[i].fd != -1; i++);
			struct fd_info* new_fds = malloc(sizeof(*new_fds) * (i + 2));
			memcpy(new_fds, fds, sizeof(*new_fds) * (i + 1));
			new_fds[i + 1].fd = -1;
			new_fds[i].fd = sock;
			new_fds[i].pid = p;
			new_fds[i].pipes[0] = pipes[0];
			new_fds[i].pipes[1] = pipes[1];
			new_fds[i].pipes2[0] = pipes2[0];
			new_fds[i].pipes2[1] = pipes2[1];
			free(fds);
			fds = new_fds;
			able_fd = true;
			kill(p, SIGUSR1);
		}
	}
}
