#!/bin/sh
# $Id: tests.sh 2 2023-03-13 04:41:08Z nishi $

ERR="[\x1b[31m\x1b[1m ERROR \x1b[0m]"
WARN="[\x1b[33m\x1b[1mWARNING\x1b[0m]"
INFO="[\x1b[33m\x1b[1m       \x1b[0m]"
SUCC="[\x1b[32m\x1b[1mSUCCESS\x1b[0m]"
PEND="[\x1b[33m\x1b[1mPENDING\x1b[0m]"
FAIL="[\x1b[31m\x1b[1m FAILED\x1b[0m]"

printf "\x1b[0m"
count=0
list=`ls -d src/tests/*.c`
for i in $list; do
	count=$(($count + 1))
done
printf "$INFO $count test(s) found.\n"
which `gmake` 2>/dev/null 1>&2
MAKE=make
if [ "$?" -eq 0 ]; then
	MAKE=gmake
fi
printf "$INFO GNU make is \`${MAKE}\`\n"
CC=`grep "CC :=" GNUmakefile | cut -d" " -f3`
printf "$INFO CC is \`${CC}\`\n"
printf "$INFO Starting the build\n"
${MAKE} all -j >/dev/null
if [ "$?" -ne 0 ]; then
	printf "$FAIL LIBRARY BUILD FAIL! exiting.\n"
	exit 1
else
	printf "$SUCC LIBRARY BUILD SUCCESS!\n"
fi
count=0
for i in $list; do
	DATE=`date +"%s"`
	NAME="${DATE}-${count}"
	DESC=`cat $i | head -n2 | tail -n1 | sed -E "s/\/\*|\*\///g" | sed -E "s/^ | $//g"`
	printf "$INFO $i is \`$DESC\`\n"
	printf "$PEND $i: Compiling"
	$CC -o /tmp/${NAME} -L`pwd`/lib -Wl,-R`pwd`/lib -I./include -lmtknife $i >/dev/null 2>log.err
	if [ "$?" -ne 0 ]; then
		printf "\r\x1b[0J$FAIL $i: Compilation Failed! See log.err.\n"
		exit 1
	else
		printf "\r\x1b[0J$SUCC $i: Compilation Success!\n"
	fi
	printf "$PEND $i: Testing"
	/tmp/${NAME} >/dev/null 2>log.err
	if [ "$?" -ne 0 ]; then
		printf "\r\x1b[0J$FAIL $i: Testing Failed! See log.err.\n"
		exit 1
	else
		printf "\r\x1b[0J$SUCC $i: Testing Success!\n"
	fi
	rm /tmp/${NAME}
	rm *.core 2>/dev/null 1>&2
	count=$(($count + 1))
done
